// This file was generated by gir (https://github.com/gtk-rs/gir @ 8b9d0bb)
// from gir-files (https://github.com/gtk-rs/gir-files @ 77d1f70)
// DO NOT EDIT

#[allow(unused_imports)]
use auto::*;

pub type Glyph = u32;
pub type LayoutRun = GlyphItem;
